@@ -1,4 +1,4 @@
-// Copyright 2018-2020 Cargill Incorporated
+// Copyright 2018-2021 Cargill Incorporated
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -15,9 +15,10 @@
 use std::collections::{HashMap, HashSet, VecDeque};
 use std::convert::TryFrom;
 use std::fmt;
-use std::path::Path;
+use std::fs;
+use std::path::{Path, PathBuf};
 use std::sync::{
-    mpsc::{channel, Receiver, Sender, TryRecvError},
+    mpsc::{channel, Receiver, RecvTimeoutError, Sender},
     Arc, RwLock,
 };
 use std::time::{Duration, Instant, SystemTime};
@@ -58,7 +59,7 @@ use super::error::{ScabbardStateError, StateSubscriberError};
 const EXECUTION_TIMEOUT: u64 = 300; // five minutes
 const CURRENT_STATE_ROOT_INDEX: &str = "current_state_root";
 const ITER_CACHE_SIZE: usize = 64;
-const COMPLETED_BATCH_INFO_ITER_RETRY: Duration = Duration::from_millis(100);
+const COMPLETED_BATCH_INFO_ITER_RETRY_MILLIS: u64 = 100;
 const DEFAULT_BATCH_HISTORY_SIZE: usize = 100;
 
 /// Iterator over entries in a Scabbard service's state
@@ -73,6 +74,8 @@ pub struct ScabbardState {
     pending_changes: Option<(String, Vec<TransactionReceipt>)>,
     event_subscribers: Vec<Box<dyn StateSubscriber>>,
     batch_history: BatchHistory,
+    state_db_file: PathBuf,
+    receipt_db_file: PathBuf,
 }
 
 impl ScabbardState {
@@ -86,8 +89,12 @@ impl ScabbardState {
         // Initialize the database
         let mut indexes = INDEXES.to_vec();
         indexes.push(CURRENT_STATE_ROOT_INDEX);
+        let state_db_file = state_db_path.to_path_buf();
+        let receipt_db_file = receipt_db_path.to_path_buf();
+        let state_db_path = state_db_path.with_extension("lmdb");
+        let receipt_db_path = receipt_db_path.with_extension("lmdb");
         let db = Box::new(LmdbDatabase::new(
-            LmdbContext::new(state_db_path, indexes.len(), Some(state_db_size))?,
+            LmdbContext::new(&state_db_path, indexes.len(), Some(state_db_size))?,
             &indexes,
         )?);
 
@@ -137,6 +144,9 @@ impl ScabbardState {
             .start()
             .map_err(|err| ScabbardStateError(format!("failed to start executor: {}", err)))?;
 
+        // initialize committed_batches metric
+        counter!("splinter.scabbard.committed_batches", 0);
+
         Ok(ScabbardState {
             db,
             context_manager,
@@ -144,13 +154,15 @@ impl ScabbardState {
             current_state_root,
             transaction_receipt_store: Arc::new(RwLock::new(TransactionReceiptStore::new(
                 Box::new(
-                    LmdbOrderedStore::new(receipt_db_path, Some(receipt_db_size))
+                    LmdbOrderedStore::new(&receipt_db_path, Some(receipt_db_size))
                         .map_err(|err| ScabbardStateError(err.to_string()))?,
                 ),
             ))),
             pending_changes: None,
             event_subscribers: vec![],
             batch_history: BatchHistory::new(),
+            state_db_file,
+            receipt_db_file,
         })
     }
 
@@ -243,8 +255,6 @@ impl ScabbardState {
         // Get the results and shutdown the scheduler
         let recv_result = result_rx.recv_timeout(Duration::from_secs(EXECUTION_TIMEOUT));
 
-        scheduler.shutdown();
-
         let batch_result = recv_result
             .map_err(|_| ScabbardStateError("failed to receive result in reasonable time".into()))?
             .ok_or_else(|| ScabbardStateError("no result returned from executor".into()))?;
@@ -252,7 +262,7 @@ impl ScabbardState {
         let batch_status = batch_result.clone().into();
         let signature = batch.batch().header_signature();
         self.batch_history
-            .update_batch_status(signature, batch_status);
+            .update_batch_status(&signature, batch_status);
 
         let txn_receipts = batch_result
             .receipts
@@ -325,7 +335,7 @@ impl ScabbardState {
                 }
 
                 self.batch_history.commit(&signature);
-
+                counter!("splinter.scabbard.committed_batches", 1);
                 Ok(())
             }
             None => Err(ScabbardStateError("no pending changes to commit".into())),
@@ -359,6 +369,29 @@ impl ScabbardState {
     pub fn clear_subscribers(&mut self) {
         self.event_subscribers.clear();
     }
+
+    /// Computes the files associated with the LMDB state, including lock files.
+    pub fn remove_db_files(&self) -> Result<(), ScabbardStateError> {
+        let state_db_path = self.state_db_file.with_extension("lmdb");
+        let state_db_lock_file_path = self.state_db_file.with_extension("lmdb-lock");
+        let receipt_db_path = self.receipt_db_file.with_extension("lmdb");
+        let receipt_db_lock_file_path = self.receipt_db_file.with_extension("lmdb-lock");
+
+        fs::remove_file(receipt_db_path.as_path()).map_err(|err| {
+            ScabbardStateError(format!("Unable to remove state LMDB file: {}", err))
+        })?;
+        fs::remove_file(receipt_db_lock_file_path.as_path()).map_err(|err| {
+            ScabbardStateError(format!("Unable to remove state LMDB file: {}", err))
+        })?;
+        fs::remove_file(state_db_path.as_path()).map_err(|err| {
+            ScabbardStateError(format!("Unable to remove receipt store LMDB file: {}", err))
+        })?;
+        fs::remove_file(state_db_lock_file_path.as_path()).map_err(|err| {
+            ScabbardStateError(format!("Unable to remove state LMDB file: {}", err))
+        })?;
+
+        Ok(())
+    }
 }
 
 fn receipts_into_transact_state_changes(
@@ -634,7 +667,7 @@ impl BatchHistory {
         let batch_info = self.upsert_batch(signature.into(), status);
 
         match batch_info.status {
-            BatchStatus::Invalid(_) | BatchStatus::Committed(_) => {
+            BatchStatus::Invalid(_) | BatchStatus::Valid(_) => {
                 self.send_completed_batch_info_to_subscribers(batch_info)
             }
             _ => {}
@@ -645,7 +678,7 @@ impl BatchHistory {
         match self.history.get_mut(signature) {
             Some(info) => match info.status.clone() {
                 BatchStatus::Valid(txns) => {
-                    self.update_batch_status(signature, BatchStatus::Committed(txns));
+                    info.set_status(BatchStatus::Committed(txns));
                 }
                 _ => {
                     error!(
@@ -756,15 +789,9 @@ impl BatchHistory {
             .batch_subscribers
             .drain(..)
             .filter_map(|(mut pending_signatures, sender)| {
-                match info.status {
-                    BatchStatus::Invalid(_) | BatchStatus::Committed(_) => {
-                        if pending_signatures.remove(&info.id) && sender.send(info.clone()).is_err()
-                        {
-                            // Receiver has been dropped
-                            return None;
-                        }
-                    }
-                    _ => (),
+                if pending_signatures.remove(&info.id) && sender.send(info.clone()).is_err() {
+                    // Receiver has been dropped
+                    return None;
                 }
 
                 if pending_signatures.is_empty() {
@@ -802,12 +829,14 @@ impl ChannelBatchInfoIter {
         timeout: Duration,
         pending_ids: HashSet<String>,
     ) -> Result<Self, ScabbardStateError> {
+        let timeout = Instant::now()
+            .checked_add(timeout)
+            .ok_or_else(|| ScabbardStateError("failed to schedule timeout".into()))?;
+
         Ok(Self {
             receiver,
-            retry_interval: std::cmp::min(timeout, COMPLETED_BATCH_INFO_ITER_RETRY),
-            timeout: Instant::now()
-                .checked_add(timeout)
-                .ok_or_else(|| ScabbardStateError("failed to schedule timeout".into()))?,
+            retry_interval: Duration::from_millis(COMPLETED_BATCH_INFO_ITER_RETRY_MILLIS),
+            timeout,
             pending_ids,
         })
     }
@@ -822,26 +851,21 @@ impl Iterator for ChannelBatchInfoIter {
             if self.pending_ids.is_empty() {
                 return None;
             }
-
-            match self.receiver.try_recv() {
-                Ok(batch_info) => match batch_info.status {
-                    BatchStatus::Invalid(_) | BatchStatus::Committed(_) => {
-                        self.pending_ids.remove(&batch_info.id);
-                        return Some(Ok(batch_info));
-                    }
-                    _ => {}
-                },
-                Err(TryRecvError::Empty) => {
-                    // Check if the timeout has expired
-                    if Instant::now() >= self.timeout {
-                        return Some(Err(format!(
-                            "timeout expired while waiting for incompleted batches: {:?}",
-                            self.pending_ids
-                        )));
-                    }
-                    std::thread::sleep(self.retry_interval);
+            // Check if the timeout has expired
+            if Instant::now() >= self.timeout {
+                return Some(Err(format!(
+                    "timeout expired while waiting for incompleted batches: {:?}",
+                    self.pending_ids
+                )));
+            }
+            // Check for the next BatchInfo
+            match self.receiver.recv_timeout(self.retry_interval) {
+                Ok(batch_info) => {
+                    self.pending_ids.remove(&batch_info.id);
+                    return Some(Ok(batch_info));
                 }
-                Err(TryRecvError::Disconnected) => return None,
+                Err(RecvTimeoutError::Timeout) => {}
+                Err(RecvTimeoutError::Disconnected) => return None,
             }
         }
     }
@@ -853,6 +877,7 @@ mod tests {
 
     use std::path::PathBuf;
 
+    use cylinder::{secp256k1::Secp256k1Context, Context};
     use tempdir::TempDir;
     use transact::{
         families::command::make_command_transaction,
@@ -860,107 +885,10 @@ mod tests {
             batch::BatchBuilder,
             command::{BytesEntry, Command, SetState},
         },
-        signing::hash::HashSigner,
     };
 
     const TEMP_DB_SIZE: usize = 1 << 30; // 1024 ** 3
 
-    /// Verify that the ChannelBatchInfoIter returns an error if no results are returned before the
-    /// timeout.
-    #[test]
-    fn channel_batch_iter_no_resuls_before_timeout() -> Result<(), Box<dyn std::error::Error>> {
-        let (_tx, rx) = channel();
-
-        let mut iter = ChannelBatchInfoIter::new(
-            rx,
-            Duration::from_secs(0),
-            vec!["batch-id-1".to_string()].into_iter().collect(),
-        )?;
-
-        assert!(iter.next().unwrap().is_err());
-
-        Ok(())
-    }
-
-    /// Verify that the ChannelBatchInfoIter returns a value if it is sent before next is called.
-    #[test]
-    fn channel_batch_iter_batch_info_before_next() -> Result<(), Box<dyn std::error::Error>> {
-        let (tx, rx) = channel();
-
-        let mut iter = ChannelBatchInfoIter::new(
-            rx,
-            Duration::from_secs(0),
-            vec!["batch-id-1".to_string()].into_iter().collect(),
-        )?;
-
-        tx.send(BatchInfo {
-            id: "batch-id-1".into(),
-            status: BatchStatus::Committed(vec![ValidTransaction::new("ab".into())]),
-            timestamp: SystemTime::now(),
-        })?;
-
-        let info = iter.next().transpose()?;
-
-        assert!(info.is_some());
-
-        let info = info.unwrap();
-
-        assert_eq!(&info.id, "batch-id-1");
-
-        let info = iter.next();
-
-        assert!(info.is_none());
-
-        Ok(())
-    }
-
-    /// Verify that the ChannelBatchInfoIter returns a value if it is sent after next is called.
-    #[test]
-    fn channel_batch_iter_batch_info_after_next() -> Result<(), Box<dyn std::error::Error>> {
-        let (tx, rx) = channel();
-
-        let barrier = Arc::new(std::sync::Barrier::new(2));
-
-        let iter_barrier = Arc::clone(&barrier);
-        let jh = std::thread::spawn(move || {
-            let mut iter = ChannelBatchInfoIter::new(
-                rx,
-                Duration::from_secs(1),
-                vec!["batch-id-1".to_string()].into_iter().collect(),
-            )
-            .unwrap();
-
-            iter_barrier.wait();
-
-            let info = iter.next().transpose().unwrap();
-
-            assert!(info.is_some());
-
-            let info = info.unwrap();
-
-            assert_eq!(&info.id, "batch-id-1");
-
-            let info = iter.next();
-
-            assert!(info.is_none());
-        });
-
-        barrier.wait();
-
-        // Wait a tiny amount for the iter.next() call.
-        std::thread::sleep(Duration::from_millis(10));
-
-        tx.send(BatchInfo {
-            id: "batch-id-1".into(),
-            status: BatchStatus::Committed(vec![ValidTransaction::new("ab".into())]),
-            timestamp: SystemTime::now(),
-        })?;
-
-        jh.join().unwrap();
-
-        Ok(())
-    }
-
     /// Verify that an empty receipt store returns an empty iterator
     #[test]
     fn empty_event_iterator() {
@@ -1051,16 +979,21 @@ mod tests {
         let address = "abcdef".to_string();
         let value = b"value".to_vec();
 
-        let signer = HashSigner::default();
+        let signing_context = Secp256k1Context::new();
+        let signer = signing_context.new_signer(signing_context.new_random_private_key());
         let batch = BatchBuilder::new()
             .with_transactions(vec![
-                make_command_transaction(&[Command::SetState(SetState::new(vec![
-                    BytesEntry::new(address.clone(), value.clone()),
-                ]))])
+                make_command_transaction(
+                    &[Command::SetState(SetState::new(vec![BytesEntry::new(
+                        address.clone(),
+                        value.clone(),
+                    )]))],
+                    &*signer,
+                )
                 .take()
                 .0,
             ])
-            .build_pair(&signer)
+            .build_pair(&*signer)
             .expect("Failed to build batch");
         state
             .prepare_change(batch)
@@ -1118,18 +1051,22 @@ mod tests {
         let address3 = "0123456789".to_string();
         let value3 = b"value3".to_vec();
 
-        let signer = HashSigner::default();
+        let signing_context = Secp256k1Context::new();
+        let signer = signing_context.new_signer(signing_context.new_random_private_key());
         let batch = BatchBuilder::new()
             .with_transactions(vec![
-                make_command_transaction(&[Command::SetState(SetState::new(vec![
-                    BytesEntry::new(address1.clone(), value1.clone()),
-                    BytesEntry::new(address2.clone(), value2.clone()),
-                    BytesEntry::new(address3.clone(), value3.clone()),
-                ]))])
+                make_command_transaction(
+                    &[Command::SetState(SetState::new(vec![
+                        BytesEntry::new(address1.clone(), value1.clone()),
+                        BytesEntry::new(address2.clone(), value2.clone()),
+                        BytesEntry::new(address3.clone(), value3.clone()),
+                    ]))],
+                    &*signer,
+                )
                 .take()
                 .0,
             ])
-            .build_pair(&signer)
+            .build_pair(&*signer)
             .expect("Failed to build batch");
         state
             .prepare_change(batch)
